/**==========================================================================

	PC-8031 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2024 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	last modified  Apr,27,2024.

;==========================================================================*/
#include "../PC-8001/PC-8001.h"
#ifdef __cplusplus
extern "C"
{
#endif
#define export
static const char* err_size = "Described size exceeds actual image length\n";
static const char* err_head = "Header size is shorter than specified\n";
static const char* err_trkz = "Track data exceeds image length\n";
static const char* err_sect = "Sector data exceeds image length\n";
static const char* err_scnt = "Amount of Sectors is over 32\n";
static const char* err_id_c = "wrong cylinder id\n";
static const char* err_id_h = "wrong head id\n";
static const char* err_id_r = "wrong sector id\n";
/**--------------------------------------------------------------------------
 **		METHODS
 **------------------------------------------------------------------------*/
static USHORT read_short(LPBYTE ptr) {
	USHORT v = 0;
	v = (USHORT)ptr[1];
	v <<= 8;
	v += (USHORT)ptr[0];
	return v;
}
static DWORD read_quad(LPBYTE ptr) {
	DWORD v = 0;
	v = (DWORD)ptr[3];
	v <<= 8;
	v += (DWORD)ptr[2];
	v <<= 8;
	v += (DWORD)ptr[1];
	v <<= 8;
	v += (DWORD)ptr[0];
	return v;
}

static int _d88_sector_count(LPD88SECTOR self) {
	return (int)read_short(self->sectors);
}

static int _d88_sector_length(LPD88SECTOR self) {
	return (int)read_short(self->length);
}

static BOOL _media_check_image(LPFLOPPYDISK self) {
	self->header = (LPD88HEADER)self->image_raw_data;
	// 記載されているｻｲｽﾞが実際のｲﾒｰｼﾞ長を超過している場合はｴﾗｰ
	if (read_quad(self->header->disksize) > self->image_length) {
		self->err_msg = err_size;
		return FALSE;
	}
	// ﾍｯﾀﾞｰ領域がｲﾒｰｼﾞ長を超過している場合はｴﾗｰ
	if (self->image_length <= sizeof(D88HEADER)) {
		self->err_msg = err_head;
		return FALSE;
	}
	memset(&self->offsets, 0xff, sizeof(self->offsets));
	// ﾄﾗｯｸﾁｪｯｸ
	for (int t = 0, t2 = 0; t < 164; ++t) {
		BYTE C = (BYTE)(t / 2);	// cylinder
		BYTE H = (BYTE)(t % 2);	// head

		// 倍ﾄﾗｯｸのﾃﾞｨｽｸでは1ｼﾘﾝﾀﾞｰごとにｽｷｯﾌﾟさせる
		if (self->header->disktype == 0x00)
			t2 = t;		// 2D
		else if ((t % 4) > 2)
			continue;	// 2DD/2HDのｼﾘﾝﾀﾞｰを1つ飛ばしにする
		else {			// 飛ばした分詰める
			C = t / 4;
			t2 = ((t / 4) * 2) + (t % 2);
		}
		DWORD offset = read_quad(&self->header->trackptr[t * 4]);
		if (offset == 0)
			continue;
		// ﾄﾗｯｸｵﾌｾｯﾄがD88ﾍｯﾀﾞｰ領域にある場合はｴﾗｰ
		if (offset < 32)
			continue;
		// ｾｸﾀﾍｯﾀﾞｰ領域がｲﾒｰｼﾞ長を超過している場合はｴﾗｰ
		if ((offset + 16) > self->image_length) {
			self->err_msg = err_trkz;
#if _DEBUG
			n80device.logger(self->err_msg);
#endif
			continue;
		}
		LPD88SECTOR track = (LPD88SECTOR)&self->image_raw_data[offset];
		USHORT sectors = read_short(track->sectors);
		if (sectors > 32) {
			self->err_msg = err_scnt;
#if _DEBUG
			n80device.logger(self->err_msg);
#endif
			continue;
		}
#if _DEBUG
		n80device.logger("track offsets[%d](C:%d/H:%d) : ", t, track->id.C, track->id.H);
#endif
		for (int s = 0; s < sectors; ++s) {
			BYTE R = (BYTE)(s + 1);
			USHORT sectlen = read_short(track->length);
			// ｾｸﾀﾃﾞｰﾀ領域がｲﾒｰｼﾞ長を超過している場合はｴﾗｰ
			if ((offset + sectlen + 16) > self->image_length) {
				self->err_msg = err_sect;
#if _DEBUG
				n80device.logger(self->err_msg);
#endif
				break;
			}
			if (track->id.C != C) {
				self->err_msg = err_id_c;
#if _DEBUG
				//n80device.logger(self->err_msg);
#endif
				//return FALSE;
			}
			if (track->id.H != H) {
				self->err_msg = err_id_h;
#if _DEBUG
				//n80device.logger(self->err_msg);
#endif
				//return FALSE;
			}
			if (track->id.R != R) {
				self->err_msg = err_id_r;
#if _DEBUG
				//n80device.logger(self->err_msg);
#endif
				//return FALSE;
			}
			self->offsets[t2][s] = offset;
#if _DEBUG
			n80device.logger("[%d:{%d,%d,%d,%d}],", s, track->id.C, track->id.H, track->id.R, track->id.N);
#endif
			offset += sectlen + 16;
			track = (LPD88SECTOR)&self->image_raw_data[offset];
		}
#if _DEBUG
		n80device.logger("\n");
#endif
	}
	return TRUE;
}

static int _d88_find_track_index(LPFLOPPYDISK self, int cylinder, BYTE HD) {
	int t = cylinder * 2 + HD;
	if (t >= 0 && t < 164)
		return t;
	return -1;
}

static LPD88SECTOR_EX _d88_enum_track_sector(LPFLOPPYDISK self, int t, int s) {
	LPD88SECTOR_EX sector = NULL;
	if (t >= 0 && t < 164 && s >= 0 && s <= 32)
		sector = self->get_sector(self, t * 100 + s);
	return sector;
}

static LPD88SECTOR_EX _d88_get_sector(LPFLOPPYDISK self, int i) {
	if (i >= 0) {
		int t = i / 100;
		int s = i % 100;
		if (t >= 0 && t < 164 && s >= 0 && s < 32) {
			DWORD offset = self->offsets[t][s];
			if (offset != 0xffffffff)
				return (LPD88SECTOR_EX)&self->image_raw_data[offset];
		}
	}
	return NULL;
}

static LPD88SECTOR_EX _d88_find_sector_in_track(LPFLOPPYDISK self, int cylinder, BYTE HD, BYTE MT, LPSECTORID id, int *track, int *sectno, int *header) {
	int t = cylinder * 2 + HD;
	for (int h = 0; h < 2; ++h) {
		if (t >= 0 && t < 164) {
			*track = t;
			*header = HD;
			for (int s = 0; s < 32; ++s) {
				DWORD offset = self->offsets[t][s];
				if (offset != 0xffffffff) {
					LPD88SECTOR sector = (LPD88SECTOR)&self->image_raw_data[offset];
					*sectno = s;
					if (sector->id.C == id->C && sector->id.H == id->H && sector->id.R == id->R) {
						return (LPD88SECTOR_EX)sector;
					}
				} else {
					break;
				}
			}
			if (HD == 0 && MT == 1) {
				++t;
				HD = 1;
				*sectno = -1;
			} else
				break;
		}
	}
	return NULL;
}
/**--------------------------------------------------------------------------
 **		Floppy Disk Initializer
 **------------------------------------------------------------------------*/
static BOOL _n88_init_media(LPFLOPPYDISK self) {
	self->find_track_index		= _d88_find_track_index;
	self->enum_track_sector		= _d88_enum_track_sector;
	self->get_sector			= _d88_get_sector;
	self->find_sector_in_track	= _d88_find_sector_in_track;
	self->sectors				= _d88_sector_count;
	self->length				= _d88_sector_length;
	return _media_check_image(self);
}
/**--------------------------------------------------------------------------
 **		Disk Image Initial Scan
 **------------------------------------------------------------------------*/
static BOOL _mgr_image_scan(LPDISKIMAGE image) {
	// clear floppy array
	image->disk_count = 0;
	for (int i = 0; i < 10; ++i)
		image->disks[i] = NULL;
	// scan floppy disks
	for (DWORD base_offset = 0, bottom_offset = 0; base_offset < image->image_length; base_offset = bottom_offset) {
		LPD88HEADER header = (LPD88HEADER)&image->image_raw_data[base_offset];
		int length = read_quad(header->disksize);
		bottom_offset = base_offset + length;
		if (image->disk_count < 10) {
			LPFLOPPYDISK disk = image->alloc();
			if (disk == NULL)
				break;
			image->disks[image->disk_count++] = disk;
			disk->image				= image;
			disk->base_offset		= base_offset;
			disk->bottom_offset		= bottom_offset;
			disk->image_length		= length;
			disk->image_raw_data	= (LPBYTE)header;
			disk->header			= header;
			disk->unit_no			= -1;
			if (!_n88_init_media(disk))
				break;
		}
	}
	return image->disk_count > 0;
}
static BOOL _mgr_detach(LPDISKIMGMGR self, int slot) {
	if (slot < 0 || slot > 1)
		return FALSE;
	LPDISKIMAGE image = self->slots[slot];
	if (image == NULL)
		return TRUE;
	for (int i = 0; i < image->disk_count; ++i) {
		int unit_no = image->disks[i]->unit_no;
		if (unit_no != -1) {
			LPFDUDRIVE drv = &self->owner->fdd.drives[unit_no];
			drv->eject(drv);
		}
	}
	self->slots[slot] = NULL;
	image->ref_count = 0;
	image->dispose(image);
	return TRUE;
}
static BOOL _mgr_attach(LPDISKIMGMGR self, LPDISKIMAGE image, int slot_no, int unit_no) {
	if (!self->scan(image))
		return FALSE;
	slot_no = slot_no == -1 ? 0 : slot_no;
	unit_no = unit_no == -1 ? (slot_no == 0 ? 0 : 1) : unit_no;
	if (slot_no < 0 || slot_no > 1 || unit_no < 0 || unit_no > 1)
		return FALSE;
	if (!self->detach(self, slot_no))
		return FALSE;

	BOOL		result = FALSE;
	LPFDUDRIVES fdd    = &self->owner->fdd;
	if (image->disk_count > 1 && unit_no == 0) {
		fdd->eject(fdd, 0);
		result = fdd->insert(fdd, 0, image->disks[0]);
		if (result) {
			fdd->eject(fdd, 1);
			result = fdd->insert(fdd, 1, image->disks[1]);
			if (!result)
				fdd->eject(fdd, 0);
		}
	} else {
		fdd->eject(fdd, unit_no);
		result = fdd->insert(fdd, unit_no, image->disks[0]);
	}
	if (result)
		self->slots[slot_no] = image;
	return result;
}
static void _mgr_dispose(LPDISKIMGMGR self) {
	for (int i = 0; i < self->maxslots; ++i) {
		self->detach(self, i);
		self->slots[i] = NULL;
	}
}
/**--------------------------------------------------------------------------
 **		Initializer
 **------------------------------------------------------------------------*/
export void _fdu_init_manager(LPDISKIMGMGR self, LPFDU fdu) {
	self->owner		= fdu;
	self->scan		= _mgr_image_scan;
	self->attach	= _mgr_attach;
	self->detach	= _mgr_detach;
	self->dispose	= _mgr_dispose;
	self->maxslots	= 2;
	for (int i = 0; i < self->maxslots; ++i)
		self->slots[i] = NULL;
}
#ifdef __cplusplus
}	// extern "C"
#endif
