/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		alsa sound
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <math.h>
#include <limits.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"

#define	ALSA_PCM_NEW_HW_PARAMS_API
#include <alsa/asoundlib.h>
#define	exportk
#define MAXSTsARRAYS	(8000)
#define	WAVETOP			(4000)
#define WAVEBOTTOM		(-4000)
#define MAXBLKARRAYS    (80)
/**--------------------------------------------------------------------------
 **	declare structures
 **------------------------------------------------------------------------*/
typedef struct _wave_T_conv {
	int         tone;       // wave tone (Hz)
	int         T;          // state count of one wave (T)
	int         halfT;      // half T
	int         frames;     // frames of one wave in a sampling rate
	int         halfs[2];   // lo/hi wave frames
	int         arr[MAXBLKARRAYS];
	int         offset;     // adjust wave phase
}   WAVETsCONV, * LPWAVETsCONV;
typedef struct _sound_ticks_tag {
	double		ticks;		// T-states累計
	DWORD		state;		// 0:off 1:on, PCGの場合はサンプル数(周波数)
}	SAMPLETs, *LPSAMPLETs;
typedef struct _sound_ticks_array_tag {
	SAMPLETs	arr[MAXSTsARRAYS];
	double		ticks_sum;	// sum of ticks
	int			index;		// current arr index
	int			count;		// number of arr factors
}	SAMPLETsARR, *LPSAMPLETsARR;
typedef struct _sound_bank_tag {
	SAMPLETs	current;
	SAMPLETsARR	banks[2];
	int			bank;
}	SAMPLETsBANK, *LPSAMPLETsBANK;
/**--------------------------------------------------------------------------
 **	external variable/function
 **------------------------------------------------------------------------*/
extern N80LX	n80lx;
extern void		select_snd_dev(void);
/**--------------------------------------------------------------------------
 **	local variables
 **------------------------------------------------------------------------*/
static WAVETsCONV	beep_wave;
static SAMPLETsBANK	beep_bank;
static SAMPLETsBANK	sound_bank[3];
/**--------------------------------------------------------------------------
 **	function prototype declare
 **------------------------------------------------------------------------*/
static void calc_wave_frame(LPWAVETsCONV, int, int);
/**--------------------------------------------------------------------------
 **	global functions
 **------------------------------------------------------------------------*/
export void dispose_alsa() {
	snd_pcm_drain(n80lx.handle);
	snd_pcm_close(n80lx.handle);
	free(n80lx.period_buffer);
}

static void listdev(PALSADEVs devs) {
	char** hints;

	int err = snd_device_name_hint(-1, "pcm", (void***)&hints);
	if (err != 0) {
		fprintf(stderr, "err: can\'t get device name hints\n");
	}
	for (char** n = hints; *n != NULL; ++n) {
		char *io = snd_device_name_get_hint(*n, "IOID");
		if (io != NULL && strcmp("Output", io) == 0) {
			char *name = snd_device_name_get_hint(*n, "NAME");
			if (name != NULL) {
				char *desc = snd_device_name_get_hint(*n, "DESC");

				if (devs->cnt == 0) {
					devs->list = calloc(1, sizeof(ALSADEV));
				} else {
					devs->list = realloc(devs->list, sizeof(ALSADEV) * (devs->cnt + 1));
				}
				if (devs->list == NULL) {
					fprintf(stderr, "err: can\'t allocate dev name buffer\n");
					break;
				}
				strncpy(devs->list[devs->cnt].name, name, 80 - 1);
				strncpy(devs->list[devs->cnt].dev, desc == NULL ? "" : desc, 80 - 1);
				devs->cnt++;
				free(name);
				if (desc != NULL)	free(desc);
			}
		}
		if (io != NULL)	free(io);
	}
	snd_device_name_free_hint((void**)hints);
}

static int open_alsa(const char *dev) {
	int	rc;

	// bank init
	memset(&beep_bank, 0, sizeof(beep_bank));
	memset(&sound_bank, 0, sizeof(sound_bank));

	// PCMデバイスを プレイバックモード で開く
	rc = snd_pcm_open(&n80lx.handle, dev, SND_PCM_STREAM_PLAYBACK, 0);
	if (rc < 0) {
		fprintf(stderr, "unable to open pcm device :%s\n", snd_strerror(rc));
		return 1;
	}
	// ハードウェアパラメータの領域を確保
	snd_pcm_hw_params_alloca(&n80lx.params);

	// n80lx.handle ハードウェアパラメータのデフォルト値を取得
	snd_pcm_hw_params_any(n80lx.handle, n80lx.params);

	// ハードウェアパラメータに Interleaved mode に設定
	rc = snd_pcm_hw_params_set_access(n80lx.handle, n80lx.params, SND_PCM_ACCESS_RW_INTERLEAVED);
	if (rc < 0) {
		fprintf(stderr, "unable to set non-interreaved :%s\n", snd_strerror(rc));
		goto exit_function;
	}

	// フォーマットは １チャネル 16bit リトルエンディアン
	rc = snd_pcm_hw_params_set_format(n80lx.handle, n80lx.params, SND_PCM_FORMAT_S16_LE);
	if (rc < 0) {
		fprintf(stderr, "unable to set format :%s\n", snd_strerror(rc));
		goto exit_function;
	}

	// ビットレートは CDのサンプリングレート (44100 bit/s)
	//int val = 44100;
	int val = 48000;
	int dir = 0;
	rc = snd_pcm_hw_params_set_rate_near(n80lx.handle, n80lx.params, (unsigned int *)&val, &dir);
	if (rc < 0) {
		fprintf(stderr, "unable to set sampling rate :%s\n", snd_strerror(rc));
		goto exit_function;
	}
	fprintf(stderr, "exam sampling rate = %d\n", val);
	n80lx.sampling_rate = val;

	// モノラル
	rc = snd_pcm_hw_params_set_channels(n80lx.handle, n80lx.params, 2);
	if (rc < 0) {
		fprintf(stderr, "unable to set mono :%s\n", snd_strerror(rc));
		goto exit_function;
	}

	// フレームサイズ
	snd_pcm_uframes_t frames = n80lx.sampling_rate / 50;	// 20ms分のフレーム数
	snd_pcm_uframes_t frames2 = frames;
	dir = 0;
	rc = snd_pcm_hw_params_set_period_size_near(n80lx.handle, n80lx.params, &frames2, &dir);
	if (rc < 0) {
		fprintf(stderr, "unable to set period size :%s\n", snd_strerror(rc));
		goto exit_function;
	}
	fprintf(stderr, "exam period size = %ld[%ld]\n", frames2, frames);
	n80lx.period_frames = frames2;

	// 区間(Period)数設定
	int periods = 4;
	rc = snd_pcm_hw_params_set_periods(n80lx.handle, n80lx.params, periods, 0);
	if (rc < 0) {
		fprintf(stderr, "unable to set number of periods :%s\n", snd_strerror(rc));
		goto exit_function;
	}

	// ハードウェアパラメータをデバイスに適用
	rc = snd_pcm_hw_params(n80lx.handle, n80lx.params);
	if (rc < 0) {
		fprintf(stderr, "unable to set hw parameters: %s\n", snd_strerror(rc));
		goto exit_function;
	}

	// 1フレーム間隔を取得するための必要なバッファサイズを取得
	dir = 0;
	snd_pcm_hw_params_get_period_size(n80lx.params, &frames, &dir);
	//fprintf(stderr, "frame = %d, dir = %d\n", frames, dir);
	n80lx.period_bytes  = frames * 2 * 2;	// 1サンプルあたり2バイトでかつ，1チャンネル
	n80lx.period_buffer = (LPUSHORT) malloc(n80lx.period_bytes);
	memset(n80lx.period_buffer, 0, n80lx.period_bytes);

	n80lx.ring_buffer_frames = snd_pcm_avail_update(n80lx.handle);
	fprintf(stderr, "avail = %ld\n", n80lx.ring_buffer_frames);

	dir = 0;
	snd_pcm_hw_params_get_period_time(n80lx.params, (unsigned int *)&val, &dir);
	fprintf(stderr, "period time = %d, dir = %d\n", val, dir);
	n80lx.period_usec = val;

	// CPU Ticks変換係数: frames = (44100/3993600*Ts)
	n80lx.frame_keisu = (double) n80lx.sampling_rate / (double) n80device.param.hz;
	calc_wave_frame(&beep_wave, n80lx.sampling_rate, 2400);

	snd_pcm_prepare(n80lx.handle);
	return 0;

exit_function:
	snd_pcm_drain(n80lx.handle);
	snd_pcm_close(n80lx.handle);
	free(n80lx.period_buffer);
	return 1;
}

static int select_alsa_device(void) {
	// open last device
	if (strlen(n80lx.snd_dev_name) > 0) {
		if (open_alsa(n80lx.snd_dev_name) == 0)
			return 0;
		memset(n80lx.snd_dev_name, 0, sizeof(n80lx.snd_dev_name));
	}
	char dev[80];
	// select snd device
	memset(&n80lx.snddevs, 0, sizeof(ALSADEVs));
	listdev(&n80lx.snddevs);
	if (n80lx.snddevs.cnt == 0)
		return -1;
	select_snd_dev();
	if (n80lx.snddevs.selected >= 0) {
		PALSADEV d = &n80lx.snddevs.list[n80lx.snddevs.selected];
		sprintf(dev, "%s", d->name);
		fprintf(stdout, "open alsa playback device \"%s\"\n", d->name);
		if (open_alsa(dev) == 0) {
			strcpy(n80lx.snd_dev_name, dev);
			return 0;
		}
	}
	return -1;
}

export int init_alsa(void) {
	if (select_alsa_device() == 0) {
		n80lx.ini.flush(&n80lx.ini);
		return 0;
	}
	return -1;
}
/**--------------------------------------------------------------------------
 **	local functions
 **------------------------------------------------------------------------*/
static inline void append_tick_to_array(LPSAMPLETsBANK bank, DWORD new_state) {
	// 現在のTick累積を配列に追加する
	LPSAMPLETsARR arr = &bank->banks[bank->bank];
	if (arr->count < MAXSTsARRAYS - 1) {
		arr->arr[arr->count].ticks   = bank->current.ticks;
		arr->arr[arr->count++].state = bank->current.state;
		arr->ticks_sum += bank->current.ticks;
	} else
		XTRACE("beep ticks array overflow\n");

	bank->current.ticks = 0;
	bank->current.state = new_state;
}

static void _n80_beep_on() {
	if (beep_bank.current.state == 0) {
		append_tick_to_array(&beep_bank, 1);
	}
}

static void _n80_beep_off() {
	if (beep_bank.current.state == 1) {
		append_tick_to_array(&beep_bank, 0);
	}
}

static void _n80_append_sound_ticks(DWORD T) {
	// BEEP:	CPU STATE数を累積する
	beep_bank.current.ticks += T;
	// PCG(PSG):CPU STATE数をALSA frame数に変換して累積する
	sound_bank[0].current.ticks += ((double)T * n80lx.frame_keisu);
	sound_bank[1].current.ticks += ((double)T * n80lx.frame_keisu);
	sound_bank[2].current.ticks += ((double)T * n80lx.frame_keisu);
}

static void _n80_sound_on(USHORT freq, int ch) {
	//PCG(PSG)全ﾁｬﾝﾈﾙの要素数を合わせる
	if (sound_bank[ch].current.state != freq) {
		for (int i = 0; i < 3; ++i) {
			USHORT f = ch == i ? freq : sound_bank[i].current.state;
			append_tick_to_array(&sound_bank[i], f);
		}
	}
}

static void _n80_sound_off() {
	append_tick_to_array(&sound_bank[0], 0);
	append_tick_to_array(&sound_bank[1], 0);
	append_tick_to_array(&sound_bank[2], 0);
}

static void _n80_clear_pcg_ticks() {
	memset(&sound_bank, 0, sizeof(sound_bank));
}

static inline int16_t *sample_overlay(int16_t *pcm, int val) {
	int32_t v = *pcm;
	v += val;
	if (v > SHRT_MAX)
		v = SHRT_MAX;
	else if (v < SHRT_MIN)
		v = SHRT_MIN;
	*pcm++ = (int16_t)v;
	*pcm++ = (int16_t)v;
	return pcm;
}
/**--------------------------------------------------------------------------
 **		Beep Maker
 **------------------------------------------------------------------------*/
static void calc_wave_frame(LPWAVETsCONV wave, int sampling_rate, int tone_hz) {
	if (tone_hz == 0) {     // lo-tone min (silent)
		wave->tone = tone_hz;
		wave->T = 0;
		wave->halfT = 0;
		wave->frames = 0;
		wave->halfs[0] = 0;
		wave->halfs[1] = 0;
	} else {
		wave->tone = tone_hz;
		wave->T = n80device.param.hz / tone_hz;
		wave->halfT = wave->T / 2;
		wave->frames = sampling_rate / tone_hz; // ex) 96,000Hz / 2400Hz = 40 frames
		if (wave->frames < 2) {
			wave->frames = 2;   // hi-tone max
		}
		wave->halfs[0] = wave->frames / 2;
		wave->halfs[1] = wave->halfs[0];
		if ((wave->frames % 2) > 0)
			++wave->halfs[1];
		int t = wave->T / wave->frames;
		int a = wave->T % wave->frames;
		for (int i = 0; i < wave->frames; ++i) {
			wave->arr[i] = t;
		}
		if (a > 0) {
			int odd = 0, b = 0;
			for (int i = 0; i < a; ++i) {
				if (b < wave->frames) {
					++wave->arr[b];
				}
				else {
					odd = odd == 0 ? 1 : 0;
					b = odd;
					++wave->arr[b];
				}
				b += 2;
			}
		}
	}
	wave->offset = 0;
}

static inline int iterator_beep_switch(LPSAMPLETsARR arr) {
	for (int i = arr->index; i < arr->count; ++i) {
		if (arr->arr[i].ticks > 0) {
			arr->arr[i].ticks--;
			arr->index = i;
			return arr->arr[i].state;
		}
	}
	return 0;	// beep off
}

static inline int _n80_beep_pcm_writer(LPSAMPLETsARR arr, int16_t *pcm) {
	int frames = (int)floor((arr->ticks_sum * n80lx.frame_keisu + 0.5));
#if _VERBOSE
	//fprintf(stderr, "beep tick = %g, k = %g,  frames = %d\n", arr->ticks_sum, n80lx.frame_keisu, frames);
#endif
	if (frames > n80lx.period_frames) {
		frames = n80lx.period_frames;
	}
	arr->index = 0;
	for (int i = 0; i < frames; ++i) {
		double volume = 0;
		if (beep_wave.offset >= beep_wave.frames) {
			beep_wave.offset = 0;
		}
		for (int t = 0; t < beep_wave.arr[beep_wave.offset]; ++t) {
			DWORD on_off = iterator_beep_switch(arr);
			if (on_off != 0)
				volume += 1.0f;
			else
				volume -= 1.0f;
		}
		if (beep_wave.offset < beep_wave.halfs[0]) {	// lo-wave always silent
			*pcm++ = WAVEBOTTOM;
			*pcm++ = WAVEBOTTOM;
		} else {
			volume /= (double)beep_wave.arr[beep_wave.offset];
			volume *= (double)WAVETOP;
			*pcm++ = (short)volume;
			*pcm++ = (short)volume;
		}
		++beep_wave.offset;
	}
	return frames;
}

static inline int _n80_sound_pcm_writer(LPSAMPLETsARR arr0, LPSAMPLETsARR arr1, LPSAMPLETsARR arr2, int16_t	*pcm) {
	int	write_frames = 0;
	double rate = n80lx.period_frames < arr0->ticks_sum ? n80lx.period_frames / arr0->ticks_sum : 1.0;
	for (int i = 0; i < arr0->count; ++i) {
		int frames = (int) floor((arr0->arr[i].ticks * rate) + 0.5);
		int remain = frames;
		if (frames > 0) {
			int wave_frames0 = arr0->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr0->arr[i].state + 0.5) : 0;
			int wave_frames1 = arr1->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr1->arr[i].state + 0.5) : 0;
			int wave_frames2 = arr2->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr2->arr[i].state + 0.5) : 0;
			int half_wave_frames0 = wave_frames0 / 2;
			int half_wave_frames1 = wave_frames1 / 2;
			int half_wave_frames2 = wave_frames2 / 2;
			for (int j = 0; j < frames; ++j) {
				if (write_frames++ < n80lx.period_frames) {
					--remain;
					int volume = 0;
					if (wave_frames0 > 0)	{	// ch.#0 sound on
						if ((n80lx.phase_adjuster % wave_frames0) < half_wave_frames0)
							volume += WAVETOP;
						else
							volume += WAVEBOTTOM;
					}
					if (wave_frames1 > 0)	{	// ch.#1 sound on
						if ((n80lx.phase_adjuster % wave_frames1) < half_wave_frames1)
							volume += WAVETOP;
						else
							volume += WAVEBOTTOM;
					}
					if (wave_frames2 > 0)	{	// ch.#2 sound on
						if ((n80lx.phase_adjuster % wave_frames2) < half_wave_frames2)
							volume += WAVETOP;
						else
							volume += WAVEBOTTOM;
					}
					pcm = sample_overlay(pcm, volume);
					if (++n80lx.phase_adjuster > 10000000)
						n80lx.phase_adjuster = 0;
				} else
					return n80lx.period_frames;
			}
		}
	}
	return write_frames;
}

static int _n80_make_pcm() {
	// rack of T states
	if (beep_bank.banks[beep_bank.bank].ticks_sum < n80device.param.sync_ticks) {
		append_tick_to_array(&beep_bank, beep_bank.current.state);
	}
#if _VERBOSE
	//int diff = (int)beep_bank.banks[beep_bank.bank].ticks_sum - (int)n80device.param.sync_ticks;
	//fprintf(stderr, "beep avail = %d(%d)T\n", (int)beep_bank.banks[beep_bank.bank].ticks_sum, diff);
#endif
	// rack of frames
	for (int ch = 0; ch < 3; ++ch) {
		if (sound_bank[ch].banks[sound_bank[ch].bank].ticks_sum < n80lx.period_frames) {
			append_tick_to_array(&sound_bank[ch], sound_bank[ch].current.state);
		}
	}
	int16_t	*pcm = (int16_t *) n80lx.period_buffer;
	int	write_frames1 = 0, write_frames2 = 0;
	memset(pcm, 0, n80lx.period_bytes);
	//---- Beep ---------------------------------------------------------------
	LPSAMPLETsARR arr1 = &beep_bank.banks[beep_bank.bank];				// current
	LPSAMPLETsARR arr2 = &beep_bank.banks[beep_bank.bank == 0 ? 1 : 0];	// next
	arr2->count = 0;
	arr2->index = 0;
	arr2->ticks_sum = 0;

	write_frames1 = _n80_beep_pcm_writer(arr1, pcm);
	// bank exchange
	arr1->count     = 0;
	arr1->index		= 0;
	arr1->ticks_sum = 0;
	beep_bank.bank = beep_bank.bank == 0 ? 1 : 0;

	//---- PCG ----------------------------------------------------------------
	pcm = (int16_t *) n80lx.period_buffer;
	LPSAMPLETsARR arr10 = &sound_bank[0].banks[sound_bank[0].bank];					// ch.#0 current
	LPSAMPLETsARR arr20 = &sound_bank[0].banks[sound_bank[0].bank == 0 ? 1 : 0];	// ch.#0 next
	LPSAMPLETsARR arr11 = &sound_bank[1].banks[sound_bank[1].bank];					// ch.#1 current
	LPSAMPLETsARR arr21 = &sound_bank[1].banks[sound_bank[1].bank == 0 ? 1 : 0];	// ch.#1 next
	LPSAMPLETsARR arr12 = &sound_bank[2].banks[sound_bank[2].bank];					// ch.#2 current
	LPSAMPLETsARR arr22 = &sound_bank[2].banks[sound_bank[2].bank == 0 ? 1 : 0];	// ch.#2 next
	arr20->count = 0;
	arr20->ticks_sum = 0;
	arr21->count = 0;
	arr21->ticks_sum = 0;
	arr22->count = 0;
	arr22->ticks_sum = 0;

	write_frames2 = _n80_sound_pcm_writer(arr10, arr11, arr12, pcm);

	// bank exchange
	arr10->count     = 0;
	arr10->ticks_sum = 0;
	arr11->count     = 0;
	arr11->ticks_sum = 0;
	arr12->count     = 0;
	arr12->ticks_sum = 0;
	sound_bank[0].bank = sound_bank[0].bank == 0 ? 1 : 0;
	sound_bank[1].bank = sound_bank[1].bank == 0 ? 1 : 0;
	sound_bank[2].bank = sound_bank[2].bank == 0 ? 1 : 0;
	
	return write_frames1 > write_frames2 ? write_frames1 : write_frames2;
}
#if _VERBOSE
	volatile int feed_cnt = 0;
	volatile int feed_wait = 0;
	volatile int feed_fail = 0;
#endif

static void _n80_feed_pcm(int write_frames) {
	//snd_pcm_prepare(n80lx.handle);
	//snd_pcm_writei(n80lx.handle, n80lx.period_buffer, write_frames);
#if _VERBOSE
	static snd_pcm_sframes_t total_write = 0;
	static snd_pcm_sframes_t total_frame = 0;
#endif
	snd_pcm_sframes_t delay, avail, remain = write_frames, wrote = 0;
	LPUSHORT ptr = &n80lx.period_buffer[0];

	snd_pcm_prepare(n80lx.handle);
	for (volatile int i = 0; i < 100;) {
		volatile int rc = snd_pcm_avail_delay(n80lx.handle, &avail, &delay);
		if (rc < 0) {
			++i;
			snd_pcm_prepare(n80lx.handle);
			//avail = delay = 0;
#if _VERBOSE
			fprintf(stderr, "rc = %d (snd_pcm_avail_delay)\n",rc);
			++feed_fail;
#endif
			//break;
		} else if (avail > 0) {
			snd_pcm_sframes_t write = avail;
			if (avail > remain)
				write = remain;
			// write pcm data to alsa ring buffer
			snd_pcm_sframes_t result = snd_pcm_writei(n80lx.handle, ptr, write);
			remain -= result;
			wrote  += result;
			if (remain > 0) {
				ptr = &n80lx.period_buffer[wrote];
			} else
				break;
		}
#if _VERBOSE
		else
			++feed_wait;
#endif
	}
#if _VERBOSE
	total_write += wrote;
	total_frame += write_frames;
	if (++feed_cnt >= 100) {
		double ave = (double)total_write / (double)feed_cnt;
		double ave2 = (double)total_frame / (double)feed_cnt;
		fprintf(stderr, "pcm write frames = %g(/%g)\n", ave, ave2);
		total_write = 0;
		total_frame = 0;

		if (feed_fail > 0 || feed_wait > 0) {
			double cnt = (double)feed_cnt;
			fprintf(stderr, "feed fail = %.3f(%%), wait = %.3f(%%)\n", ((double)feed_fail / cnt) * 100, ((double)feed_wait / cnt) * 100);
		}
		feed_cnt = feed_fail = feed_wait = 0;
	};
#endif
}

export void _n80_init_snd() {
	n80device.snd.beep_on		= _n80_beep_on;
	n80device.snd.beep_off		= _n80_beep_off;
	n80device.snd.sound_on		= _n80_sound_on;
	n80device.snd.sound_off		= _n80_sound_off;
	n80device.snd.append_ticks	= _n80_append_sound_ticks;
	n80device.snd.clear_ticks	= _n80_clear_pcg_ticks;
	n80device.snd.make_pcm		= _n80_make_pcm;
	n80device.snd.feed_pcm		= _n80_feed_pcm;
}