/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2021 H.Kanemoto
 **
 **		bitmap font(bdf) file access
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"
#define	export

typedef enum {
	CHARS = 0,
	ENC = 1,
	BMP = 2,
	BIN = 3,
}	SCANMODE;
/**--------------------------------------------------------------------------
 **	external variable/functions
 **------------------------------------------------------------------------*/
extern N80LX		n80lx;
extern N80DEVICE	n80device;
/**--------------------------------------------------------------------------
 **	local functions (methods)
 **------------------------------------------------------------------------*/
static void bdf_dispose(PBDF self) {
	if (self->bitmap != NULL) {
		free(self->bitmap);
		self->bitmap = NULL;
		self->cnt	 = 0;
	}
}
static uint16_t bdf_sjis2jis(uint16_t sjis) {
	int h = sjis >> 8;
	int l = sjis & 0xff;
   
	if (h <= 0x9f) {
	  if (l < 0x9f)
		h = (h << 1) - 0xe1;
	  else
		h = (h << 1) - 0xe0;
	} else {
	  if (l < 0x9f)
		h = (h << 1) - 0x161;
	  else
		h = (h << 1) - 0x160;
	}
	if (l < 0x7f)
	  l -= 0x1f;
	else if (l < 0x9f)
	  l -= 0x20;
	else
	  l -= 0x7e;
	return (uint16_t)(h << 8 | l);
}

static uint16_t bdf_jis2sjis(uint16_t jis) {
	int h = jis >> 8;
	int l = jis & 0xff;

	if (h & 1) {
		if (l < 0x60)
		l += 0x1f;
		else
		l += 0x20;
	} else {
		l += 0x7e;
	}
	if (h < 0x5f)
		h = (h + 0xe1) >> 1;
	else
		h = (h + 0x161) >> 1;
	return (uint16_t)(h << 8 | l);
}

static PBDFBMP bdf_find(PBDF self, uint16_t sjis) {
	if (self->cnt > 0 && self->bitmap != NULL) {
		uint16_t jis = self->sjis2jis(sjis);
		for (int i = 0; i < self->cnt; ++i) {
			if (self->bitmap[i].jis == jis)
				return &self->bitmap[i];
		}
	}
	return NULL;
}

static bool	bdf_load(PBDF self) {
	char		buff[512];
	int			chars  = 0;
	int			index  = 0;
	int			bmpcnt = 0;
	uint32_t	bit;
	bool		loop = true;

	FILE *fp = fopen(self->path, "r");
	if (fp == NULL)
		return false;
	SCANMODE mode = CHARS;
	while (loop) {
		char 	*line;
		size_t	size = 0;
		ssize_t	rsiz;
		rsiz = getline(&line, &size, fp);
		if (rsiz >= 0) {
			switch (mode) {
				case CHARS:
					if (strstr(line, "CHARS ") != NULL) {
						if (sscanf(line, "%s %d", buff, &chars) == 2) {
							if (chars > 0) {
								self->bitmap = calloc(chars, sizeof(BDFBMP));
								if (self->bitmap == NULL)
									loop = false;
								else
									mode = ENC;
							}
						}
					}
					break;
				case ENC:
					if (strstr(line, "STARTCHAR ") != NULL) {
						uint32_t jis;
						if (sscanf(line, "%s %x", buff, &jis) == 2) {
							self->bitmap[index].jis = (u_int16_t)jis;
							mode = BMP;
						}
					}
					break;
				case BMP:
					if (strstr(line, "BITMAP") != NULL) {
						bmpcnt = 0;
						mode = BIN;
					}
					break;
				case BIN:
					if (sscanf(line, "%x", &bit) == 1) {
						self->bitmap[index].bitmap[0][bmpcnt] = (BYTE)((bit >> 8) & 0xff);
						self->bitmap[index].bitmap[1][bmpcnt] = (BYTE)(bit& 0xff);
#ifdef _VERBOSE
						//for (int i = 0; i < 16; ++i) {
						//	fprintf(stderr, "%c", bit & 0x8000 ? '*' : '-');
						//	bit <<= 1;
						//}
						//fprintf(stderr, "\n");
#endif
						if (++bmpcnt > 15) {
							if (++self->cnt < chars) {
								mode = ENC;
								++index;
							} else
								loop = false;
						}
					} else
						mode = ENC;
					break;
			}
		} else
			loop = false;
		free(line);
	}
	fprintf(stderr, "bdf font load %d/%d\n", self->cnt, chars);
	return true;
}
/**--------------------------------------------------------------------------
 **		Initializer
 **------------------------------------------------------------------------*/
export bool init_bdf(PBDF self, const char *bdfname) {
	bool result = true;

	memset(self, 0, sizeof(BDF));
	self->dispose	= bdf_dispose;
	self->sjis2jis	= bdf_sjis2jis;
	self->jis2sjis	= bdf_jis2sjis;
	self->load		= bdf_load;
	self->find		= bdf_find;

	memset(self->path, 0, sizeof(self->path));
	strncpy(self->path, n80lx.home, sizeof(n80lx.home));
	strcat(self->path, "/");
	strcat(self->path, bdfname);
	fprintf(stderr, "bdf path = %s\n", self->path);

	if (!self->load(self)) {
		self->cnt = 0;
		result = false;
	}
	return result;
}
