/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Zaurus(Spitz)
 **		--------------------------------------------
 **		Copyright (c) 1993-2021 H.Kanemoto
 **
 **		sound device selector
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	2010/12/08	Zaurus SL-C3000 Fedora Core 10 ARM版 porting
 **		2.00	2017/03/27	Raspberry Pi2 Raspbian(wheezy) ARM版 porting
 **------------------------------------------------------------------------*/
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <locale.h>
#include <sys/ioctl.h>
//#include <sys/mman.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include "../include/n80.h"
#include "../pc-8001/pc-8001.h"

#define export
/**--------------------------------------------------------------------------
 **	external value/function
 **------------------------------------------------------------------------*/
extern N80LX	n80lx;
extern void		draw_icon(LPDWORD video, int pitch, int x, int y);
extern void		draw_text(int line, LPBYTE attr, LPBYTE text);
/**--------------------------------------------------------------------------
 **	internal functions
 **------------------------------------------------------------------------*/
static void show() {
	int r;
	for (r = 0; r < 25; ++r)
		draw_text(r, (LPBYTE)&n80lx.attr[r][0], (LPBYTE)&n80lx.text[r][0]);
	draw_icon(n80lx.video, 640, 12, 16);
	SDL_UpdateTexture(n80lx.texture, NULL, n80lx.video, 640 * 4);
	SDL_RenderClear(n80lx.renderer);
	SDL_RenderCopy(n80lx.renderer, n80lx.texture, NULL, NULL);
	SDL_RenderPresent(n80lx.renderer);
}

static void drawitem(PALSADEV item, int line, bool selected) {
	int row = line + 9;
	if (item != NULL) {	// show item row
		for (int i = 1, c = 0; i < 79; ++i) {
			if (item->name[c] != '\0')
				n80lx.text[row][i] = item->name[c++];
			else
				n80lx.text[row][i] = '\0';
		}
		for (int i = 1; i < 79; ++i) {
			n80lx.attr[row][i] = selected ? 0x22 : 0x07;
		}
		if (selected) {
			char ch;
			for (int i = 8, c = 0; i < 79; ++i) {
				if ((ch = item->dev[c]) != '\0') {
					n80lx.text[7][i] = item->dev[c++];
					if (ch < 0x20)
						n80lx.text[7][i] = '\0';
				} else
				n80lx.text[7][i] = '\0';
			}
		}
	} else {			// show blank row
		for (int i = 1; i < 79; ++i) {
			n80lx.text[row][i] = '\0';
		}
		for (int i = 1; i < 79; ++i) {
			n80lx.attr[row][i] = 0x07;
		}
	}
}
/**--------------------------------------------------------------------------
 **	LX_alsa_devs methods
 **------------------------------------------------------------------------*/
static int devs_row_to_index(PALSADEVs self) {
	return self->page_start_index(self) + self->cur_row;
}
static int devs_page_start_index(PALSADEVs self) {
	return self->cur_page * self->rows_per_page;
}
static int devs_page_end_index(PALSADEVs self) {
	int s = self->page_start_index(self);
	int r = self->rows_in_page(self, self->cur_page);
	return s + r - 1;
}
static int devs_rows_in_page(PALSADEVs self, int page /* 0 start */) {
	if (page < self->pages)
		return self->rack_page == page ? self->rack_page_rows : self->rows_per_page;
	else
		return 0;
}
static void devs_show_page(PALSADEVs self) {
	int s = self->page_start_index(self);
	int e = self->page_end_index(self);
	for (int r = 0; r < 15; ++r) {
		if (s <= e)
			drawitem(&self->list[s++], r, r == self->cur_row);
		else
			drawitem(NULL, r, false);
	}
}
static bool devs_page_down(PALSADEVs self) {
	if (self->cur_page == (self->pages - 1))
		return false;	// 最終頁では何もしない
	//次の頁に移動
	++self->cur_page;
	self->cur_row = 0;
	self->rows = self->rows_in_page(self, self->cur_page);
	self->show_page(self);
	return true;
}
static bool devs_page_up(PALSADEVs self) {
	if (self->cur_page > 0) {
		//前の頁に移動
		--self->cur_page;
		self->rows = self->rows_in_page(self, self->cur_page);
		self->cur_row = self->rows - 1;
		self->show_page(self);
		return true;
	} else
		return false;	// 先頭頁では何もしない
}
static bool devs_down(PALSADEVs self) {
	if (self->cur_row == (self->rows - 1)) {
		//最終行では次頁への移動を試みる
		return self->pgdn(self);
	}
	drawitem(&self->list[self->row_to_index(self)], self->cur_row, false);
	++self->cur_row;
	drawitem(&self->list[self->row_to_index(self)], self->cur_row, true);
	return true;
}
static bool devs_up(PALSADEVs self) {
	if (self->cur_row > 0) {
		drawitem(&self->list[self->row_to_index(self)], self->cur_row, false);
		--self->cur_row;
		drawitem(&self->list[self->row_to_index(self)], self->cur_row, true);
		return true;
	} else	//先頭行では前頁への移動を試みる
		return self->pgup(self);
}
static int devs_init(PALSADEVs self, int rows_per_page) {
	self->rows_in_page		= devs_rows_in_page;
	self->page_start_index	= devs_page_start_index;
	self->page_end_index	= devs_page_end_index;
	self->row_to_index		= devs_row_to_index;
	self->show_page			= devs_show_page;
	self->down				= devs_down;
	self->up				= devs_up;
	self->pgdn				= devs_page_down;
	self->pgup				= devs_page_up;

	self->cur_page = 0;
	self->selected = -1;
	if (self->cnt > 0) {
		self->rows_per_page = rows_per_page;
		self->pages	= self->cnt / self->rows_per_page;
		self->rack_page_rows = self->cnt % self->rows_per_page;
		if (self->rack_page_rows > 0) {
			self->rack_page = self->pages;
			++self->pages;
		} else
			self->rack_page = -1;	// full house
		self->rows = self->rows_in_page(self, self->cur_page);
		self->cur_row = 0;
	} else
		self->pages = 0;
	return 0;
}
/**--------------------------------------------------------------------------
 **	global function
 **------------------------------------------------------------------------*/
export void select_snd_dev(void) {
	PALSADEVs devs = &n80lx.snddevs;
	devs_init(devs, 15);
	if (devs->pages == 0)
		return;

	for (int r = 0; r < 25; ++r) {
		for (int c = 0; c < 80; ++c) {
            BYTE atr = 0x23;
            if ((r == 4 || r == 5) && (c > 0 && c < 79))
                atr = 0x27;
            else if (r == 7 && (c > 0 && c < 79))
                atr = 0x26;
            else if ((r > 8 && r < 24) && (c > 0 && c < 79))
                atr = 0x07;
			n80lx.attr[r][c] = atr;
			n80lx.text[r][c] = 32;
		}
	}
	const char *title1 = "    \xea\xeaN80\xea\xea PC-8001 Emulator for Linux";
	const char *title2 = "    Copyright 1993-2025 (C) H.Kanemoto, All Right Reserved.";
	const char *title3 = "Select Alsa Sound Device";
	const char *title4 = "usage:  [ESC]...Quit [\x1e\x1f]...Move  [PgUp/PgDn]...Page [RET]...Select";
	const char *title5 = "hint: select sound playback device listed below:";
	memcpy(&n80lx.text[1][2], title1, strlen(title1));
	memcpy(&n80lx.text[2][2], title2, strlen(title2));
	memcpy(&n80lx.text[4][2], title3, strlen(title3));
	memcpy(&n80lx.text[5][2], title4, strlen(title4));
	memcpy(&n80lx.text[7][2], title5, strlen(title5));
	devs->show_page(devs);
    show();

	SDL_Event	event;
	bool		shift = false;
	for (int quit = 0; !quit;) {
		while (SDL_PollEvent(&event)) {
			if (event.type == SDL_QUIT)
				quit = 1;
			else if (event.type == SDL_KEYUP) {
				if (event.key.keysym.sym == SDLK_LSHIFT || event.key.keysym.sym == SDLK_RSHIFT)
					shift = false;
			} else if (event.type == SDL_KEYDOWN) {
				if (event.key.keysym.sym == SDLK_LSHIFT || event.key.keysym.sym == SDLK_RSHIFT)
					shift = true;
				else if (event.key.keysym.sym == SDLK_SPACE || event.key.keysym.sym == SDLK_RETURN) {
					if (event.key.repeat == 0) {
						devs->selected = devs->row_to_index(devs);
						quit = 1;	break;
					}
				} else if (event.key.keysym.sym == SDLK_ESCAPE) {
					if (event.key.repeat == 0) {
						quit = 1;	break;
					}
				} else if ((event.key.keysym.sym == SDLK_TAB && !shift) || event.key.keysym.sym == SDLK_DOWN) {
					if (devs->down(devs))
						show();
				} else if ((event.key.keysym.sym == SDLK_TAB && shift) || event.key.keysym.sym == SDLK_UP) {
					if (devs->up(devs))
						show();
				} else if (event.key.keysym.sym == SDLK_PAGEDOWN) {
					if (devs->pgdn(devs))
						show();
				} else if (event.key.keysym.sym == SDLK_PAGEUP) {
					if (devs->pgup(devs))
						show();
				}
			}
		}
	}
}
